package guru.colormatcher;

import processing.core.PImage;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * @author guru
 *
 * ColorMatcher is a tool for finding the center of a group of pixels sharing the same color.
 * the machtColors function takes a array of colors and returns a array of CMPoint objects 
 * representing the center of all pixels sharing the same color. if a color isn't found null 
 * is returned. 
 * 
 * The matching algorithm can be choosen in the contructor using the CM_RGB or CM_nRGB constatns.
 * by default the RGB machter is used.
 */

public class ColorMatcher {
    public static int CM_RGB = 0;
    public static int CM_nRGB = 1;

    private int matcher = CM_RGB;

    public ColorMatcher() {
    }
    
    public ColorMatcher( int matcher ) {
        this.matcher = matcher;
    }

    public CMPoint[] matchColors( PImage img, int[] colors ) {
        ArrayList tmp = new ArrayList();
        for (int i =0; i < colors.length; i++ ) {
            tmp.add( new ArrayList());
        }

        int idx = 0;
        for ( int y = 0; y < img.height ; y++ ) {
            for ( int x = 0; x < img.width; x++ ) {
                for (int i = 0; i < colors.length; i++ ) {
                    if( match( colors[i], img.pixels[idx] )) {
                        ((ArrayList)tmp.get(i)).add( new CMPoint( x, y ));
                    }
                }
                idx ++;
            }
        }

        CMPoint[] result = new CMPoint[tmp.size()]; 
        for (int i = 0; i < tmp.size(); i++ ) {
            if (((ArrayList)tmp.get(i)).size() > 0) {
                result[i] = avg( (ArrayList)tmp.get(i));
            } else {
                result[i] = null;
            }
        }
        return result;    
    }

    private boolean match( int c1, int c2 ) {
        if ( matcher == CM_RGB ) {
            return matchRGB( c1, c2 );
        } else if ( matcher == CM_nRGB ) {
            return matchNormalizedRGB( c1, c2 );
        } else {
            return matchRGB( c1, c2 );
        }
    }

    private boolean matchNormalizedRGB( int c1, int c2 ) {
        float limit = 0.02f;
        int r = c1 >> 16 & 0xFF;
        int g = c1 >> 8 & 0xFF;
        int b = c1 & 0xFF;

        if (r + g + b  == 0 ) return false;
        float sr = 1.0f * r / (r + g + b );
        float sg = 1.0f * g / (r + g + b );
        float sb = 1.0f * b / (r + g + b );

        r = c2 >> 16 & 0xFF;
        g = c2 >> 8 & 0xFF;
        b = c2 & 0xFF;

        if (r + g + b  == 0 ) return false;
        float cr = 1.0f * r / (r + g + b );
        float cg = 1.0f * g / (r + g + b );
        float cb = 1.0f * b / (r + g + b );

        return cr > sr - limit && cr < sr + limit &&
            cg > sg - limit && cg < sg + limit && 
            cb > sb - limit && cb < sb + limit;
    }

    boolean matchRGB( int c1, int c2 ) {
        int limit = 20;
        int sr = c1 >> 16 & 0xFF;
        int sg = c1 >> 8 & 0xFF;
        int sb = c1 & 0xFF;

        int cr = c2 >> 16 & 0xFF;
        int cg = c2 >> 8 & 0xFF;
        int cb = c2 & 0xFF;

        return cr > sr - limit && cr < sr + limit &&
            cg > sg - limit && cg < sg + limit && 
            cb > sb - limit && cb < sb + limit;
    }

    private CMPoint avg( ArrayList l ) {
        if (l.size() == 0) {
            return new CMPoint( 0, 0 );
        }
        int x = 0;
        int y = 0;
        for( Iterator i = l.iterator(); i.hasNext(); ) {
            CMPoint p = (CMPoint)i.next();
            x += p.x;
            y += p.y;
        }
        return new CMPoint( x  / l.size(), y / l.size());
    }

}

